%{
/**
 | Desktop calculator grammar
 | From the bison manual example, modified by MLO
 | $Id: dtc.y,v 1.5 2000/04/06 06:08:08 loreti Exp $
**/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "dtcdef.h"

#define YYDEBUG 0               /* 1 for bison debug info on stderr  */

static double *last;            /* Pointer to the last printed value */

static void hex(void);          /* dec -> hex/oct conversion         */
static void oct(void);

static void printTree(void);    /* Symbol table printout: tree,      */
static void pN(const symRec *); /*   and single nodes                */

static double my_asin(double);  /* Wrapper functions, that check the */
static double my_acos(double);  /*   argument before calling the     */
static double my_log(double);   /*     corresponding math library    */
static double my_log10(double); /*       procedure                   */
static double my_sqrt(double);
 
static double trunc(double);    /* Integer part, fractional part and */
static double frac(double);     /*   rounding to the nearest integer */
static double round(double);

static void prompt(void);
static int  interactive;

%}

%union {                        /* YYSTYPE, possible return values */
  double val;
  symRec *tptr;
}

%token <val>  NUMBER            /* Basic tokens: numbers, symbols, */
%token <tptr> SYMBOL            /*   functions, expressions,       */
%token <tptr> FUNCTION          /*     special conversions.        */
%token <tptr> SPECIAL
%type  <val>  exp

%right '='                      /* Operators, left- and right- */
%left  '-' '+'                  /*   associative, from low to  */
%left  '*' '/' '%'              /*   high precedence order.    */
%right '^'

%%

input:          /* empty */ | input line ;

line:           '\n' { prompt(); }                                            ;
line:           ';'                                                           ;
line:           error '\n' { prompt(); }                                      ;
line:           error ';'                                                     ;
line:           exp '\n' { printf ("  %.10g\n", (*last=$1)); prompt(); }      ;
line:           exp ';'  { printf ("  %.10g\n", (*last=$1));           }      ;
line:           SPECIAL '\n' { (*($1->value.v_funptr))(); prompt(); }         ;
line:           SPECIAL ';'  { (*($1->value.v_funptr))();           }         ;
                                                                              
exp:            NUMBER               { $$ = $1; }                             ;
exp:            '-' NUMBER           { $$ = -$2; }                            ;
exp:            SYMBOL               { $$ = $1->value.var; }                  ;
exp:            SYMBOL '=' exp       { $1->value.var = $3;  $$ = $3; }        ;
exp:            FUNCTION '(' exp ')' { $$ = (*($1->value.d_funptr))($3); }    ;
exp:            FUNCTION             { $$ = (*($1->value.d_funptr))(*last); } ;
exp:            exp '+' exp          { $$ = $1 + $3; }                        ;
exp:            exp '-' exp          { $$ = $1 - $3; }                        ;
exp:            exp '*' exp          { $$ = $1 * $3; }                        ;
exp:            exp '/' exp          { if ($3 == 0) {
                                         puts("  Error: division by zero");
                                         $$ = 0;
                                       } else {
                                         $$ = $1 / $3;
                                       }
                                     }                                        ;
exp:            exp '%' exp          { if ($3 == 0) {
                                         puts("  Error: division by zero");
                                         $$ = 0;
                                       } else {
                                         int i = $1 / $3;
                                         $$ = $1 - i * $3;
                                       }
                                     }                                        ;
exp:            exp '^' exp          { $$ = pow($1, $3); }                    ;
exp:            '(' exp ')'          { $$ = $2; }                             ;

%%

/* Main program */

static funTable funs[] = {              /* Predefined functions */
  {"sin",    sin},
  {"cos",    cos},
  {"tan",    tan},
  {"asin",   my_asin},
  {"acos",   my_acos},
  {"atan",   atan},
  {"ln",     my_log},
  {"log",    my_log10},
  {"exp",    exp},
  {"sqrt",   my_sqrt},
  {"trunc",  trunc},
  {"round",  round},
  {"frac",   frac},
  {0, 0}
};

static constTable cons[] = {            /* Predefined constants */
  {"pi",     3.1415926535897932385},
  {"e",      2.7182818284590452354},
  {"dtor",   0.01745329252},
  {"rtod",   57.29577951},
  {"last",   0},                        /* _MUST_BE_ the last one */
  {0, 0}
};

static specTable spec[] = {
  {"hex",    hex},
  {"oct",    oct},
  {"stab",   printTree},
  {0, 0}
};

symRec *root = 0;

static void initTable(void);

int main()
{
  initTable();

#if YYDEBUG != 0
  yydebug = YYDEBUG;
#endif

  interactive = isatty(STDIN_FILENO);
  prompt();
  yyparse();
  if (interactive) puts("");
  return EXIT_SUCCESS;
}

static void initTable(void)
{
  symRec *p;
  int     i;

  for (i = 0;   funs[i].name != 0;   ++i) {
    root = insertNode(root, funs[i].name, 0, &p);
    p->type           = FUNCTION;
    p->value.d_funptr = funs[i].fun;
  }

/**
 | The value field for the last constant entered is pointed to by "last";
 | *last is updates after every printed expression, and the constant
 | "last" recalls its value.
**/

  for (i = 0;   cons[i].name != 0;   ++i) {
    root = insertNode(root, cons[i].name, 0, &p);
    p->type                   = SYMBOL;
    *(last = &(p->value.var)) = cons[i].value;
  }

  for (i = 0;   spec[i].name != 0;   ++i) {
    root = insertNode(root, spec[i].name, 0, &p);
    p->type           = SPECIAL;
    p->value.v_funptr = spec[i].fun;
  }
}

symRec *insertNode(
        symRec  *pNode,
  const char    *name,
        int     *pNew,
        symRec **p
) {

/**
 | Looks recursively, in the binary tree containing the symbol table,
 | for a node containing the ASCII string "name"; "pNode" points to
 | the node we are looking at currently.  If an existing node cannot
 | be found, an empty new one is created and inserted in the tree; a
 | pointer to the target node will be stored in "*p".  If "pNew" is
 | not NULL, in *pNew will be stored or 0 (an existing node has been
 | found) or 1 (a new node has been created).
**/

  if (pNode) {
    int compare;

    if ((compare = strcmp(name, pNode->name)) == 0) {
      if (pNew) *pNew = 0;
      *p = pNode;
    } else if (compare < 0) {
      pNode->left = insertNode(pNode->left, name, pNew, p);
    } else {
      pNode->right = insertNode(pNode->right, name, pNew, p);
    }
  } else {

/**
 | Only the left- and right- pointers and the name field (using the
 | so called 'struct hack') are initialized
**/

    pNode = malloc(sizeof(symRec) + strlen(name));
    if (pNode) {
      pNode->left = pNode->right = 0;
      strcpy(pNode->name, name);
      if (pNew) *pNew = 1;
      *p = pNode;
    } else {
      fputs("No heap memory available\n", stderr);
      exit(EXIT_FAILURE);
    }
  }
  return pNode;
}

int yyerror(
  char *pc
){
  printf("  Error: %s\n", pc);
  return 0;
}

static void hex(void) {
  unsigned long dec = *last;

  printf("  %lu decimal = 0x%lX hex\n", dec, dec);
}

static void oct(void) {
  unsigned long dec = *last;

  printf("  %lu decimal = 0%lo octal\n", dec, dec);
}

static void printTree(void) {
  printf("  Current symbol table:\n");
  pN(root);
}

static void pN(
  const symRec *pNode
) {
  if (pNode->left) pN(pNode->left);

  switch (pNode->type) {
    case SYMBOL:
      printf("  - %s = %f\n", pNode->name, pNode->value.var);
      break;
    default:
      break;
  }

  if (pNode->right) pN(pNode->right);
}

static void prompt(void) {
  if (interactive) printf("? ");
}

static double trunc(
  double val
) {
  int i = val;

  return i;
}

static double frac(
  double val
) {
  int i = val;

  return fabs(val - i);
}

static double round(
  double val
) {
  int i = floor(val + 0.5);

  return i;
}

static double my_asin(
  double rhs
) {
  if (rhs > 1.0   ||   rhs < -1.0) {
    puts("  Invalid argument");
    return 0;
  } else {
    return asin(rhs);
  }
}

static double my_acos(
  double rhs
) {
  if (rhs > 1.0   ||   rhs < -1.0) {
    puts("  Invalid argument");
    return 0;
  } else {
    return acos(rhs);
  }
}

static double my_log(
  double rhs
) {
  if (rhs <= 0.0) {
    puts("  Invalid argument");
    return 0;
  } else {
    return log(rhs);
  }
}

static double my_log10(
  double rhs
) {
  if (rhs <= 0.0) {
    puts("  Invalid argument");
    return 0;
  } else {
    return log10(rhs);
  }
}

static double my_sqrt(
  double rhs
) {
  if (rhs < 0.0) {
    puts("  Invalid argument");
    return 0;
  } else {
    return sqrt(rhs);
  }
}
